# Copyright (C) 2019, 2020, 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import asyncio
import abc
import copy
import dataclasses
import logging
import re
import typing
import weakref
from ipaddress import IPv4Address
from .. import sh
from ..dependency_injection import *
from ..config import ConfigLayout
from ..utils import permute_identifier, when_needed, memoproperty, is_optional_type, get_type_args
import carthage.kvstore
from ..machine import ssh_origin, ssh_origin_vrf, Machine, AbstractMachineModel
from .config import V4Config
ssh_origin_vrf_key = ssh_origin_vrf

logger = logging.getLogger('carthage.network')

_cleanup_substitutions = [
    (re.compile(r'[-_\. ]'), ''),
    (re.compile(r'database'), 'db'),
    (re.compile(r'test'), 't'),
    (re.compile(r'router'), 'rtr'),
    (re.compile(r'\..+'), ''),
]

_allocated_interfaces = set()


def if_name(type_prefix, layout, net, host=""):
    "Produce 14 character interface names for networks and hosts"
    global _allocated_interfaces

    def cleanup(s, maxlen):
        for m, r in _cleanup_substitutions:
            s = m.sub(r, s)
        return s[0:maxlen]

    assert len(type_prefix) <= 3
    layout = cleanup(layout, 2)
    maxlen = 13 - len(layout) - len(type_prefix)
    net = cleanup(net, max(3, maxlen - len(host)))
    maxlen -= len(net)
    host = cleanup(host, maxlen)
    if host:
        host += "-"
    id = "{t}{l}{h}{n}".format(
        t=type_prefix,
        l=layout,
        n=net,
        h=host)
    for i in permute_identifier(id, 14):
        if i not in _allocated_interfaces:
            _allocated_interfaces.add(i)
            return i
    assert False  # never should be reached


@dataclasses.dataclass
class NetworkInterface:

    network: object
    ifname: str
    delete_interface: bool = True
    closed: bool = dataclasses.field(init=False, default=False)

    def __del__(self):
        self.close()

    def close(self):
        if self.closed:
            return
        if self.delete_interface:
            self.delete_networking()
        self.closed = True

    def delete_networking():
        try:
            sh.ip("link", "del", self.ifname, _bg=False)
        except sh.ErrorReturnCode:
            pass
        self.closed = True


@dataclasses.dataclass
class VlanInterface(NetworkInterface):

    vlan_id: int = 0

    def __init__(self, id, network: BridgeNetwork, **kwargs):
        super().__init__(ifname="{}.{}".format(
            network.bridge_name, id), network=network, **kwargs)
        self.vlan_id = id
        self.closed = False


class TechnologySpecificNetwork(AsyncInjectable):

    '''
    Abstract base class  for accessing a network.

    The :class:`.Network` class defines the interface to a virtual network.  However different backends require different ways of accessing a network.  For KVM we need a local bridge or macvlan interfaces.  Vmware needs some form of Portgroup on a VLAN.  This class is the abstract interface to that.

'''

    async def also_accessed_by(self, others: typing.List[TechnologySpecificNetwork]):
        '''
        Abstract method to notify a class of other technology specific networks.

        After construction, if any other technologies are in use, this method is called listing all of those technologies.  Later, if other technologies are added, this method is called again.

'''
        pass


class Network(AsyncInjectable):

    '''
    Represents a network that VMs and containers can connect to.  In Carthage, networks are identified by a name and a VLAN identifier.

    How networks are accessed depends on the underlying technology.  The base Network class maintains an :class:`.Injector` so that only one instance of a technology-specific network is made for each logical network.

    .. seealso:

        BridgeNetwork
            For `carthage.Container` and `carthage.Vm`

        VmwareNetwork
            for `carthage.vmware.Vm`

    '''

    network_links: weakref.WeakSet

    def __init__(self, name, vlan_id=None, **kwargs):
        import carthage.kvstore
        super().__init__(**kwargs)
        self.name = name
        self.vlan_id = vlan_id
        self.injector.add_provider(this_network, self)
        self.injector.add_provider(V4Pool)
        self.technology_networks = []
        self.network_links = weakref.WeakSet()

    def add_network_link(self, link:NetworkLink):
        '''Indicate that  a network link has been added.  After this function teruns *link* in in *self.network_links*.
        '''
        self.network_links.add(link)
        self.injector.emit_event(
            InjectionKey(Network), "add_link",
            self, link=link,
            adl_keys={InjectionKey(Network, name=self.name)})

    def assign_addresses(self, link:NetworkLink=None):
        '''If links have addresses assigned from a :class:`carthage.kvstore.V4Pool`, assign those addresses.

        :param link: If non-None, only assign addresses for the given link.
        '''
        import carthage.kvstore
        pool = self.injector.get_instance(V4Pool)
        if link:
            return pool.assignment_loop([link])
        pool.new_assignments()
        pool.assignment_loop(self.network_links)
        
    async def access_by(self, cls: TechnologySpecificNetwork):
        '''Request a view of *self* using *cls* as a technology-specific lens.

        :return: The instance of *cls* for accessing this network
        '''
        assert issubclass(cls, TechnologySpecificNetwork), \
            "Must request access by a subclass of TechnologySpecificNetwork"
        instance = None
        if (cls not in self.ainjector) and self.vlan_id is not None:
            try:
                instance = await self.ainjector.get_instance_async(InjectionKey(cls, vlan_id=self.vlan_id))
                self.ainjector.add_provider(instance)
            except KeyError:
                pass
        if not instance:
            instance = await self.ainjector.get_instance_async(cls)
        assert cls in self.ainjector, \
            f"It looks like {cls} was not registered with add_provider with allow_multiple set to True"
        if instance not in self.technology_networks:
            await instance.also_accessed_by(list(self.technology_networks))
            l = [instance]
            for elt in self.technology_networks:
                await elt.also_accessed_by(l)
            self.technology_networks.extend(l)
        return instance

    def close(self, canceled_futures=None):
        self.ainjector.close(canceled_futures=canceled_futures)
        self.technology_networks = []
        for l in list(self.network_links):
            try:
                l.close()
            except BaseException:
                logger.exception(f'Error closing link for {repr(self)}')


    @property
    def v4_config(self):
        '''
        The :class:`V4Config` for this network.  When assigned to a *Network*, a *V4Config* cannot have any deferred elements.
        '''
        return self._v4_config

    @v4_config.setter
    def v4_config(self, config):
        assert isinstance(config, V4Config)
        config.after_resolve()
        self._v4_config = config

    def __init_subclass__(cls, **kwargs):
        if 'v4_config' in cls.__dict__:
            config = cls.v4_config
            del cls.v4_config
            config.after_resolve()
            cls._v4_config = config
        super().__init_subclass__(**kwargs)
            

this_network = InjectionKey(Network, role="this_network")


@inject(
    config_layout=ConfigLayout,
    injector=Injector,
    net=this_network)
class BridgeNetwork(TechnologySpecificNetwork):

    def __init__(self, net, *, bridge_name=None,
                 delete_bridge=None,
                 delete_interfaces=None, **kwargs):
        super().__init__(**kwargs)
        if delete_bridge is None:
            delete_bridge = not self.config_layout.persist_local_networking
        self.delete_bridge = delete_bridge
        self.name = net.name
        self.interfaces = weakref.WeakValueDictionary()
        if bridge_name is None:
            self.bridge_name = if_name('br', self.config_layout.container_prefix, self.name)
        else:
            self.bridge_name = bridge_name
        self.closed = False
        self.members = []
        if delete_interfaces is None and delete_bridge:
            delete_interfaces = True
        if delete_interfaces is None:
            delete_interfaces = not self.config_layout.persist_local_networking
        self.delete_interfaces = delete_interfaces

    async def async_ready(self):
        try:
            sh.ip('link', 'show', self.bridge_name, _bg=False)
        except sh.ErrorReturnCode_1:
            sh.ip('link', 'add', self.bridge_name, 'type', 'bridge', _bg=False)
            sh.ip("link", "set", self.bridge_name,
                  "type", "bridge", "stp_state", "1",
                  "forward_delay", "3", _bg=False)
            sh.ip("link", "set", self.bridge_name, "up", _bg=False)
        return await super().async_ready()

    def close(self):
        if self.closed:
            return
        self.members.clear()
        if self.delete_interfaces:
            self.delete_networking()
        self.closed = True
        self.interfaces.clear()

    def delete_networking(self):
        # Copy the list because we will mutate
        for i in list(self.interfaces.values()):
            try:
                i.close()
            except BaseException:
                logger.debug("Error deleting interface {}".format(i))
        if self.delete_bridge:
            logger.info("Network {} bringing down {}".format(self.name, self.bridge_name))
            sh.ip('link', 'del', self.bridge_name, _bg=False)
            self.closed = True

    def __del__(self):
        self.close()

    def add_member(self, interface):
        sh.ip("link", "set",
              interface.ifname,
              "master", self.bridge_name, "up", _bg=False)
        # We also keep a reference so that if it is a weak interface off another object it is not GC'd
        self.members.append(interface)

    def add_veth(self, link, namespace):
        bridge_member = if_name('ci', self.config_layout.container_prefix, self.name, link.machine.name)
        args = []
        if link.mtu:
            args.extend(['mtu', link.mtu])
        args.extend(['type', 'veth', 'peer'])
        if link.mac:
            args.extend(['address', str(link.mac)])
        args.extend(['name', link.interface])
        args.extend(['netns', namespace.name])
        logger.debug('Network {} creating virtual ethernet for {}'.format(self.name, link.machine.name))
        try:
            sh.ip('link', 'add', 'dev', bridge_member,
                  *args, _bg=False)
        except sh.ErrorReturnCode_2:
            logger.warn("Network {}: {} appears to exist; deleting".format(self.name, bridge_member))
            sh.ip('link', 'del', bridge_member, _bg=False)
            sh.ip('link', 'add', 'dev', bridge_member,
                  *args)
        sh.ip('link', 'set', bridge_member, 'master', self.bridge_name, 'up', _bg=False)
        ve = VethInterface(network=self, ifname=bridge_member, internal_name=link.interface, delete_interface=False)
        self.interfaces[bridge_member] = ve
        return ve

    def expose_vlan(self, id):
        iface = VlanInterface(id, self)
        ifname = iface.ifname
        try:
            sh.ip("link", "add",
                  "link", self.bridge_name,
                  "name", ifname,
                  "type", "vlan",
                  "id", id, _bg=False)
        except sh.ErrorReturnCode_2:
            logger.warn("{} appears to already exist".format(ifname))
        self.interfaces[ifname] = iface
        return iface


@dataclasses.dataclass
class VethInterface(NetworkInterface):

    internal_name: str = ""

    def close(self):
        if self.closed:
            return
        try:
            del self.network.interfaces[self.ifname]
        except KeyError:
            pass
        super().close()


class NetworkConfig:

    '''Represents a network configuration for a :class:`~carthage.machine.Machine`.  A network config maps interface names to a network alink.  A network link contains a MAC address, a network, and other information.  Eventually a MAC is represented as a string and a
    network as a Network object.  However indirection is possible in
    two ways.  First, an injection key can be passed in; this
    dependency will be resolved in the context of an
    environment-specific injector.  Secondly, a callable can be passed
    in.  This callable will be called in the context of an injector
    and is expected to return the appropriate object.

    '''

    def __init__(self):
        self.link_specs = {}

    def add(self, interface, net, mac, **kwargs):
        assert isinstance(interface, str)
        kwargs['mac'] = mac
        kwargs['net'] = net
        NetworkLink.validate(kwargs, unresolved=True)
        self.link_specs[interface] = kwargs

    def __repr__(self):
        res = f'<{self.__class__.__name__}  {repr(self.link_specs)}>'
        return res

    @inject(ainjector=AsyncInjector)
    async def resolve(self, connection, ainjector) -> dict[str, NetworkLink]:
        '''
        Return a mapping of interfaces to NetworkLinks
        The *network_links* property of *connection* is updated based on the new network links.  That side effect is a bit unclean, but doing the update here allows :meth:`carthage.machine.Machine.resolve_networking` and :meth:`carthage.machine.AbstractMachineModel.resolve_networking` to have less duplicated code.
        '''
        async def resolve1(r: typing.any, i, args, k):
            if isinstance(r, InjectionKey):
                r = await ainjector.get_instance_async(r)
            elif callable(r):
                r = await ainjector(r, i)
            if args is not None:
                args[k] = r
            return r

        def handle_other(link, other_args, other_future):
            def callback(future):
                other_link = None
                try:
                    try:
                        other = future.result()
                    except Exception:
                        logger.exception(f'Error resolving other side of {link.interface} link of {link.machine}')
                        return
                    if not isinstance(other, (Machine, AbstractMachineModel)):
                        logger.error(
                            f'The other side of the {other_interface} link to {link.machine} must be an Machine or AbstractMachineModel, not {other}')
                        return
                    if other_interface in other.network_links:
                        other_link = other.network_links[other_interface]
                        if link.net != other_link.net:
                            logger.error(
                                f'Other side of {link.interface} link on {link.machine} connected to {other_link.net} not {link.net}')
                            return
                        if 'mac' in other_args and other_link.mac and \
                                other_link.mac != other_args['mac']:
                            logger.error(
                                f'Other side of {link.interface} link on {link.machine} has MAC {other_link.mac} not {other_args["mac"]}')
                            return
                        for k, v in other_args.items():
                            setattr(other_link, k, v)
                    else:  # other link exists
                        try:
                            other_args['net'] = link.net
                            other_link = NetworkLink(other, other_interface, other_args)
                        except Exception:
                            logger.exception(
                                f'Error constructing {other_interface} link on {other} from {link.interface} link on {link.machine}')
                            return
                    link.other = other_link
                    other_link.other = link
                    other.network_links[other_link.interface] = other_link
                    other_link.member_of = []
                    other_future.set_result(other_link)
                finally:
                    if not other_future.done():
                        other_future.set_result(None)

            other_interface = other_args.pop('other_interface')
            if 'other_member_of' in other_args:
                raise ValueError('You cannot set member_of on the other side of a link.')
            if 'other_v4_config' in other_args:
                # It's possible to implement this.  We need to make sure after_resolve gets called all the time and resolve gets called on a new v4_config.  Since this function is not async, that gets messy.
                raise NotImplementedError('Currently you cannot set other_v4_config')
            for k in list(other_args):
                k_new = k[6:]
                other_args[k_new] = other_args.pop(k)
            return callback

        d = {}
        futures = []
        other_futures = []
        # For most of the machinery we want to define link memberships
        # in one direction: links have a members attribute that points
        # to their members and the rest is just memoized computation.
        # However, when specifying a configuration it is often
        # valuable (for example when specifying vlan links) to specify
        # things in the other direction.  So, we pull member_of off
        # the arguments and handle it later.
        members_of: dict[str, list] = {}
        for i, link_spec in self.link_specs.items():
            link_args = {}
            link_args_dict = link_spec.pop('link_args', {})
            if callable(link_args_dict) or isinstance(link_args_dict, InjectionKey):
                link_args_dict = await resolve1(link_args_dict, i, None, None)
                link_spec = dict(link_spec)
            link_spec.update(link_args_dict)
            del link_args_dict
            for k, v in link_spec.items():
                if k == 'other' or k.startswith('other_'):
                    link_args.setdefault('_other', {})
                    if k != 'other' and (callable(v) or isinstance(v, InjectionKey)):
                        futures.append(asyncio.ensure_future(resolve1(v, i, link_args['_other'], k)))
                    else:
                        link_args['_other'][k] = v
                    continue
                if k == 'member_of':
                    if callable(v) or isinstance(v, InjectionKey):
                        futures.append(asyncio.ensure_future(resolve1(v, i, members_of, i)))
                    else:
                        members_of[i] = v
                    continue
                if callable(v) or isinstance(v, InjectionKey):
                    futures.append(asyncio.ensure_future(resolve1(v, i, link_args, k)))
                else:
                    link_args[k] = v
            d[i] = link_args

        await asyncio.gather(*futures)
        del futures
        result: dict[str, NetworkLink] = {}
        for i, args in d.items():
            other_args = args.pop('_other', None)
            if other_args:
                if not 'other_interface' in other_args and 'other' in other_args:
                    raise RuntimeError(
                        f'At least other_interface and other must be specified when specifying the other side of a link; {i} link on {connection}')
            try:
                link = NetworkLink(connection, i, args)
                await link.resolve(ainjector=ainjector, interface=i)
                result[i] = link
            except Exception:
                logger.exception(f"Error constructing {i} link on {connection}")
                raise
            if other_args:
                other_target = other_args.pop('other')
                # The future handling is complicated.  We want
                # something like
                # modeling.base.ModelingGroup.resolve_networking to be
                # able to wait for both sides of the link to come up.
                # But we also need a future to track when the target of
                # the other side has been resolved, because we cannot
                # construct the link before then.  So, we generate one
                # future for the other_target, and another future for
                # the fully constructed other link.
                other_future = ainjector.loop.create_future()
                other_futures.append(other_future)
                other_callback = handle_other(link, other_args, other_future)
                if isinstance(other_target, InjectionKey):
                    other_target_future = asyncio.ensure_future(ainjector.get_instance_async(other_target))
                elif callable(other_target):
                    other_target_future = asyncio.ensure_future(ainjector(other_target))
                else:
                    other_target_future = ainjector.loop.create_future()
                    other_target_future.set_result(other_target)
                # We don't wait on other_target_future because it's quite possible that
                # the other side of the link will not resolve until after we resolve.
                other_target_future.add_done_callback(other_callback)
        # Now handle member_of and turn into members
        for i, member_of in members_of.items():
            for member in member_of:
                if member not in result:
                    raise ValueError(f'{i} has {member} in member_of, but {member} is not a link on {connection}')
                if not hasattr(result[member], 'members'):
                    result[member].members = []
                result[member].members.append(i)

        for k, link in result.items():
            if k in connection.network_links:
                connection.network_links[k].close()
            connection.network_links[k] = link
        for l in connection.network_links.values():
            l.member_of = []
            try:
                del l.member_links
            except BaseException:
                pass
            try:
                del l.member_of_links
            except BaseException:
                pass
        for l in connection.network_links.values():
            l.member_links

        ainjector.emit_event(InjectionKey(NetworkConfig),
                             "resolved", self,
                             pending_futures=other_futures,
                             machine=connection,
                             )
        del other_futures
        self._handle_deferred_macs(result)

        return result

    @staticmethod
    def _handle_deferred_macs(links: dict):
        from .mac import find_mac_first_member
        for l in links.values():
            if l.mac == 'inherit':
                l.mac = find_mac_first_member(l)


class VlanList(abc.ABC):

    '''Either an int, sequence of integers, a slice, or a sequence of slices
'''
    # This would be better represented as a typing.Union, but
    # unfortunately the NetworkLink.validate code cannot cope with that
    # without digging into internals of typing.

    @staticmethod
    def canonicalize(item: VlanList, link: NetworkLink):
        result = []
        if isinstance(item, (int, Network, CollectVlansType)):
            item = [item]
        for i in item:
            if isinstance(i, CollectVlansType):
                from .switch import link_collect_vlans
                result.extend(link_collect_vlans(link))
            elif isinstance(i, Network):
                if not i.vlan_id:
                    raise ValueError(f'{i} has no vlan_id set')
                result.append(i.vlan_id)
            elif isinstance(i, slice):
                result.append(i)
            elif isinstance(i, int):
                result.append(i)
            else:
                raise ValueError(f'{i} is not a valid VlanList member')
        return tuple(result)


VlanList.register(int)
VlanList.register(list)
VlanList.register(tuple)
VlanList.register(slice)


class CollectVlansType:
    pass


VlanList.register(CollectVlansType)
collect_vlans = CollectVlansType()


@dataclasses.dataclass
class NetworkLink:

    interface: str
    mac: typing.Optional[str]
    net: Network
    other: typing.Optional[NetworkLink]
    machine: object
    mtu: typing.Optional[int]
    local_type: typing.Optional[str]
    untagged_vlan: typing.Optional[int]
    allowed_vlans: typing.Optional[VlanList]
    v4_config: typing.Optional[V4Config] = dataclasses.field(default=None, repr=False)
    lldp: typing.Optional[bool] = dataclasses.field(default=True, repr=False)
    required: typing.Optional[bool] = dataclasses.field(default=True, repr=False)
    #: If true, this interface is essential and networkd should keep it up even if a dhcp lease expires or networkd is stopped
    precious: typing.Optional[bool] = dataclasses.field(default=False, repr=False)

    admin_status: typing.Optional[str] = dataclasses.field(default='up', repr=False)
    #: Sometimes it is desirable to have a different dns entry for an
    # interface than for the host as a whole. If set to a string, this
    # is the (potentially unqualified) dns name for the interface.  If
    # set to an empty string, the interface should not be registered in
    # dns.  Whether None implies the host should be registered under its name depends on how dns is configured.
    dns_name: typing.Optional[str] = None

    #: In NAT environments it is desirable to have a public DNS name separate than a private DNS name.  If set, then  when the public IP address is known, register this in dns.  It is more likely that this name needs to be fully qualified than dns_name.
    public_dns_name: typing.Optional[str] = None

    def __new__(cls, connection, interface, args):
        if 'local_type' in args:
            cls = NetworkLink.local_type_registry[args['local_type']]
        return super().__new__(cls)

    def __init__(self, connection, interface, args):
        self.validate(args)
        self.machine = connection
        self.interface = interface
        self.__dict__.update(args)
        self.net.add_network_link(self)
        for k in NetworkLink.__annotations__:
            if not hasattr(self, k):
                setattr(self, k, None)
        if self.mtu is None and getattr(self.net, 'mtu', None):
            self.mtu = self.net.mtu
        self.member_of = []
        if self.v4_config:
            self.v4_config = copy.copy(self.v4_config)

    def __hash__(self):
        return id(self)

    def __eq__(self, other):
        return self is other

    async def instantiate(self, cls: typing.Type[TechnologySpecificNetwork]):
        if self.local_type is not None:
            self.net_instance = None
            return
        try:
            return self.net_instance
        except BaseException:
            pass
        self.net_instance = await self.net.access_by(cls)
        return self.net_instance

    async def resolve(self, ainjector, interface):
        if 'merged_v4_config' in self.__dict__:
            raise ValueError('resolve called too late.')
        if self.v4_config is not None:
            await self.v4_config.resolve(ainjector=ainjector, interface=interface)
            
    def __init_subclass__(cls, **kwargs):
        if hasattr(cls, 'local_type') and cls.local_type:
            NetworkLink.local_type_registry[cls.local_type] = cls
        super().__init_subclass__(**kwargs)

    local_type_registry: typing.ClassVar[typing.Mapping[str, NetworkLink]] = weakref.WeakValueDictionary()

    @classmethod
    def validate(cls, args: dict, unresolved: bool = False):
        try:
            subclass = NetworkLink.local_type_registry[args['local_type']]
        except KeyError:
            subclass = cls
        hints = typing.get_type_hints(subclass)
        if 'member' in args:
            args['members'] = [args['member']]
            del args['member']
        for k, t in hints.items():
            if k in ('machine', 'connection', 'interface', 'member_of'):
                if k in args:
                    raise TypeError(f'{k} cannot be specified directly')
                continue
            optional = is_optional_type(t)
            if (not optional) and t.__class__ == typing._GenericAlias:
                continue
            if optional:
                t = get_type_args(t)
            if k not in args:
                if not optional:
                    raise TypeError(f'{k} is required')
            elif (not unresolved) and (not isinstance(args[k], t)):
                raise TypeError(f'{k} must be a {t} not {args[k]}')
        if subclass:
            subclass.validate_subclass(args, unresolved=unresolved)

    def __repr__(self):
        cls = self.__class__
        result = f'<{cls.__name__} '
        try:
            hints = typing.get_type_hints(cls)
            attributes = set(hints.keys())
            attributes -= {'local_type_registry', 'v4_config'}
            attributes |= {'merged_v4_config'}
            not_in_dict = set() #Attributes that are not in self.__dict__ before being queried.
            for k in attributes:
                if k not in self.__dict__: not_in_dict.add(k)
                if k == 'net':
                    result += f"net={self.net.name}"
                elif k == 'machine':
                    result += f'machine={self.machine.name}'
                else:
                    val = getattr(self, k, None)
                    if val is not None:
                        result += ' '+k+'='+repr(val)
        except Exception:
            result += 'repr failed'
        finally:
            for k in not_in_dict & set(self.__dict__.keys()):
                del self.__dict__[k]
        result += '>'
        return result
    
                
    @classmethod
    def validate_subclass(cls, args, unresolved: bool): pass

    @memoproperty
    def member_links(self):
        res = []
        if not hasattr(self, 'members'):
            return res
        for l in self.members:
            try:
                res.append(self.machine.network_links[l])
            except KeyError:
                raise KeyError(f'{l} not found as an interface on {self.machine}') from None
        for link in res:
            if self.interface not in link.member_of:
                link.member_of.append(self.interface)
        return tuple(res)

    @memoproperty
    def member_of_links(self):
        res = []
        for l in self.member_of:
            try:
                res.append(self.machine.network_links[l])
            except KeyError:
                raise KeyError(f'{l} interface not found on {self.machine}') from None
        return tuple(res)

    def _merge(self, a):
        res = {}
        if hasattr(self.net, a):
            res.update(getattr(net, a))
        if hasattr(self, a):
            res.update(getattr(self, a))
        return res


    @memoproperty
    def merged_v4_config(self):
        '''Takes *self.v4_config* as a starting point, filling in any values specified on the network's *v4_config* that are not overridden in *self*.

        The merged config may be further modified by values discovered operationally.  As an example, if STUN or instance inspection discoveres the public address of some NAT connection, then that may be filled in by an implementation of :class:`~carthage.machine.Machine`.

        The merged configuration is guaranteed to be a (possibly shallow) copy of the link's *v4_config*.  That is changing a value directly on *self.merged_v4_config* will not change *self.v4_config*.
        

        '''
        if self.v4_config:
            merged = self.v4_config.merge(getattr(self.net, 'v4_config', None))
            if merged.address == merged.gateway:
                merged.gateway = None
            return merged
        if self.v4_config:
            return copy.copy(self.v4_config)
        elif hasattr(self.net, 'v4_config'):
            return copy.copy(self.net.v4_config)
        else: return V4Config()

    @memoproperty
    def private_to_public_map(self):
        result = {}
        def add(private, public):
            if private and public: result[private] = public
        config = self.merged_v4_config
        add(config.address, config.public_address)
        for a in config.secondary_addresses:
            add(a.private, a.public)
        return result
    

    def close(self):
        if self.net:
            try:
                self.net.network_links.remove(self)
            except BaseException:
                pass
        if self.machine:
            try:
                del self.machine.network_links[self.interface]
            except BaseException:
                pass
        other = self.other
        self.other = None
        if other:
            other.close()
        try:
            self.net_instance.close()
        except BaseException:
            pass
        self.net_instance = None
        self.machine = None
        self.net = None


@inject(config_layout=ConfigLayout)
class NetworkConfigInstance(Injectable):

    def __init__(self, entries, config_layout):
        self.config_layout = config_layout
        self.entries = entries

    def __iter__(self):
        '''Return net, interface, MAC tuples.  Note that the caller is
        responsible for making the interface names line up correctly given the
        technology in question.
        '''

        for i, v in self.entries.items():
            yield v['net'], i, v['mac']


external_network_key = InjectionKey(Network, role="external")


@inject(config_layout=ConfigLayout,
        injector=Injector)
class ExternalNetwork(Network):

    def __init__(self, config_layout, injector):
        vlan_id = config_layout.external_vlan_id
        external_bridge_name = config_layout.external_bridge_name
        kwargs = {}
        if vlan_id:
            kwargs['vlan_id'] = vlan_id
        super().__init__(name="external network", injector=injector,
                         **kwargs)
        self.ainjector.add_provider(InjectionKey(BridgeNetwork),
                                    when_needed(BridgeNetwork, bridge_name=external_bridge_name, delete_bridge=False))

    @classmethod
    def supplementary_injection_keys(cls, k):
        yield external_network_key
        yield from super().supplementary_injection_keys(k)


external_network = when_needed(ExternalNetwork)


@dataclasses.dataclass
class HostMapEntry:

    ip: str
    mac: str = None


host_map_key = InjectionKey('host_map')


@inject(host_map=host_map_key, ainjector=AsyncInjector)
def mac_from_host_map(i, host_map, ainjector):
    from .machine import Machine
    machine = ainjector.get_instance(InjectionKey(Machine, _ready=False))
    entry = host_map[machine.name]
    machine.ip_address = entry.ip
    return entry.mac


@inject(ssh_origin=ssh_origin,
        )
def access_ssh_origin(ssh_origin, ssh_origin_vrf=None, extra_nsenter=[]):
    '''
        A container can be used as an ssh_origin, using the container as
        an injection point for entering a network under test.  This is
        typically done by constructing an ``nsenter`` command to enter the
        appropriate namespaces.  This function accomplishes that.

        :return: A list of arguments to be included in a *sh* call

        :param extra_nsenter: Extra namespaces to enter; something like ``['-p', '-m']``

'''
    if ssh_origin_vrf is None:
        try:
            ssh_origin_vrf = ssh_origin.injector.get_instance(ssh_origin_vrf_key)
        except KeyError:
            pass
    vrf = []
    if ssh_origin_vrf:
        vrf = ['ip', 'vrf',
               'exec', ssh_origin_vrf]
    return sh.nsenter.bake('-t', ssh_origin.container_leader,
                           '-n',
                           *vrf)


def hash_network_links(network_links: dict[str, NetworkLink]):
    '''
    Return a hash value suitable for determining whether network_links have changed in setup_tasks.
'''
    def hash_subitem(i):
        result = 0
        if i is None:
            return 0
        if isinstance(i, int):
            return i
        for v in i:
            if isinstance(v, list):
                result += hash_subitem(v)
            elif isinstance(v, dict):
                result += hash_subitem(v.keys())
                result += hash_subitem(v.values())
            elif isinstance(v, str):
                for ch in v:
                    result += ord(ch)
            elif isinstance(v, int):
                result += v
        return result

    result = hash_subitem(network_links.keys())
    for v in network_links.values():
        result += hash_subitem(v.net.name)
        if v.mac:
            result += hash_subitem(v.mac)
        if v.machine:
            result += hash_subitem(v.machine.name)
        if v.mtu:
            result += v.mtu
        # Do not include public_v4_address because it tends to change regularly
        if v.allowed_vlans:
            result += hash_subitem(VlanList.canonicalize(v.allowed_vlans, v))
        if v.untagged_vlan:
            result += v.untagged_vlan
        if v.v4_config:
            result += hash_subitem(v.v4_config.__dict__.values())
        for attr in ('speed', 'portchannel_member', 'breakout_mode'):
            result += hash_subitem(getattr(v, attr, ''))
        try:
            result += hash_subitem(v.members)
        except AttributeError:
            pass
    return result


__all__ = r'''Network TechnologySpecificNetwork BridgeNetwork
    external_network_key HostMapEntry mac_from_host_map host_map_key
access_ssh_origin
NetworkConfig NetworkLink
VlanList collect_vlans
hash_network_links
this_network
    '''.split()
@inject_autokwargs(
    injector=Injector,
    network=this_network)
class V4Pool(carthage.kvstore.HashedRangeAssignments):

    def __init__(self, domain=None, **kwargs):
        network = kwargs['network']
        if domain is None:
            domain = network.name+'/v4_pool'
        super().__init__(domain=domain, **kwargs)
        self.network = network
        self.network.injector.add_event_listener(InjectionKey(carthage.Network), 'add_link', self._invalidate_caches_cb)
        # It is not guaranteed that all models will be instantiated
        # prior to address assignment, but in the common cases such as
        # CarthageLayout's async_ready calling layout level
        # resolve_networking, that will be the case.  The consequences
        # if we get this wrong are limited if prefer_reallocation is
        # False (the default).  We will prefer to use up all the
        # addresses before reusing them.  It would be possible to more
        # accurately set this setting, for example by looking up
        # CarthageLayout and triggering on resolve_networking or
        # similar.  Doing so would require a dependency on the
        # modeling layer (or some abstract base class to use) and a
        # way to do downward-directed events, which is not easy in
        # January of 2023.
        self.enable_key_validation()

    def close(self):
        try:
            self.network.remove_event_listener(InjectionKey(Network), self._invalidate_caches_cb)
        except BaseException: pass

    def _invalidate_caches_cb(self, *args, **kwargs):
        try: del self.valid_keys
        except AttributeError: pass

    @memoproperty
    def valid_keys(self):
        keys = set()
        for l in self.network.network_links:
            keys.add(self.link_key(l))
        return frozenset(keys)

    def find_bounds(self, link):
        v4_config = link.merged_v4_config
        if v4_config is None: return
        if v4_config.pool is None: return None
        return v4_config.pool

    def record_assignment(self, key, obj, assignment):
        if obj.v4_config is None:
            obj.v4_config = carthage.network.V4Config()
        obj.v4_config.address = IPv4Address(assignment)
        try: del obj.merged_v4_config
        except AttributeError: pass

    def link_key(self, link):
        return f'{link.machine.name}|{link.interface}'

    def assignment_loop(self, links):
        for link in links:
            bounds = self.find_bounds(link)
            if not bounds: continue
            key = self.link_key(link)
            if link.v4_config and link.v4_config.address:
                self.force_assignment(key, link, link.v4_config.address)
            else:
                self._assign(key, link)
    def str_to_assignment(self, assignment):
        return IPv4Address(assignment)
    
                
__all__ += ['V4Pool']

def match_link(links: dict[str,NetworkLink], 
               interface, *,
               mac=None, net:Network=None, address=None,
               excluded_links=None):
    '''Attempt to find the :class:`NetworkLink` corresponding to an interface on a VM.
    Ideally, links can be matched by name.  However, not all :class:`~carthage.machine.Machine` implementations will preserve link names in all situations.

    The following matches are tried in order:

    #. A link with interface of *interface* and compatible *net* and *mac*.  

    #. A link with mac address of *mac* and compatible *net*.

    #. A link with address of *address* and compatible *net* and *mac*.

    In the above, compatible means that either one property is None or the two properties are equal.

    :return: A matching link from *links* or None.

    :param excluded_links: A set of interface names that will not be
    matched.  It is an error if *interface* is in *excluded_links*.
    If a matching link is found and *excluded_links* is specified, it
    will be added to *excluded_links*.  That way, when this function
    is used in a a loop, links are matched at most once.
    '''
    def compatible(a,b):
        if a is None or b is None: return True
        return a == b
    if address: address = IPv4Address(address)
    if excluded_links is None: excluded_links = set()
    assert interface not in excluded_links
    match = None
    # try explicit name match
    try: match = links[interface]
    except KeyError: pass
    if match:
        if not (compatible(match.mac, mac) and compatible(match.net, net)):
            match = None
    # if we did not find a match,  enumerate if we have a mac address
    if mac is not None and match is None:
        for match in links.values():
            if match.interface in excluded_links: continue
            if match.mac != mac: continue
            if compatible(match.net, net): break
        else: match = None
    # Now try address matching
    if match is None and address is not None:
        for match in links.values():
            if match.interface in excluded_links: continue
            if match.merged_v4_config.address != address: continue
            if compatible(mac, match.mac) and compatible(net, match.net): break
        else: match = None
    if match:
        excluded_links.add(match.interface)
    logger.debug("Matching link %s, mac %s, net %s, address %s: %s",
                 interface, mac, net, address, match)
    return match

__all__ += ['match_link']


from . import links as network_links




# Copyright (C) 2019, 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import os
import os.path
import pytest
from carthage.dependency_injection import *
from carthage.pytest import *
from carthage.vmware import *
from carthage import ConfigLayout
from pyVmomi import vim
from carthage.network import NetworkConfig, external_network_key


@pytest.fixture()
@async_test
@inject(config=ConfigLayout)
async def config(config):
    vmc = config.vmware
    for k in ('hostname', 'password', 'username', 'folder', 'cluster'):
        if getattr(vmc, k, None) is None:
            pytest.skip("Vmware Carthage is inadequately configured")
    return config


@pytest.fixture()
def vm_folder(ainjector, loop, config):
    f1 = loop.run_until_complete(ainjector(VmFolder))
    f2 = loop.run_until_complete(ainjector(VmFolder, parent=f1, name="test_vmware"))
    yield f2
    loop.run_until_complete(f2.delete())


@async_test
async def test_vm_create(vm_folder, ainjector):
    v = await ainjector(Vm, parent=vm_folder, name="blah")


@async_test
async def test_clone_vm(ainjector, vm_folder):
    v = await ainjector(VmTemplate, disk=None, parent=vm_folder, name="blah2", template=None)
    v2 = await ainjector(Vm, template=v, parent=vm_folder, name="clone1")


@async_test
async def test_clone_increase_disk_size(ainjector, vm_folder):
    class VmBigDisk(Vm):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.disk_size = self.disk_size * 2
            self.template_snapshot = None

    v = await ainjector(VmTemplate, disk=None, parent=vm_folder, name="blah3", template=None)
    v2 = await ainjector(VmBigDisk, template=v, parent=vm_folder, name="clone1big")
    d = 0
    for dev in v2.mob.config.hardware.device:
        if isinstance(dev, vim.vm.device.VirtualDisk):
            d = dev.capacityInBytes
            break
    assert d == v2.disk_size
    v3 = await ainjector(Vm, template=v2, name="clone of clone", parent=vm_folder)


@async_test
async def test_vm_with_network(ainjector, vm_folder):
    net_config = await ainjector(NetworkConfig)
    net_config.add('eth0', external_network_key, None)
    v = await ainjector(VmTemplate, template=None, disk=None, name="blah-net1", network_config=net_config)
    v2 = await ainjector(Vm, template=v, name="blah-net-clone", network_config=net_config)



# Copyright (C) 2018, 2019, 2020, 2021, 2022, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import asyncio
import contextlib
import dataclasses
import io
import os
import time
from pathlib import Path
from .dependency_injection import inject, AsyncInjector, Injector, AsyncInjectable, Injectable, InjectionKey, dependency_quote, is_obj_ready
from .config import ConfigLayout
from .setup_tasks import SetupTaskMixin, setup_task
from . import sh, machine
from .utils import memoproperty, when_needed
from pathlib import Path


@dataclasses.dataclass
class RsyncPath:

    machine: machine.Machine
    path: str

    def __repr__(self):
        return f'<Rsync {self.machine}:{self.path}>'

    def __str__(self):
        return f'{self.machine.ip_address}:{self.path}'

    @property
    def relpath(self):
        '''Path relative to the root directory of machine'''
        p = Path(self.path)
        if p.is_absolute():
            return str(p.relative_to("/"))
        else:
            return self.path

    @memoproperty
    def ssh_origin(self):
        from .machine import ssh_origin
        try:
            return self.machine.injector.get_instance(ssh_origin)
        except KeyError:
            return None

    @memoproperty
    def ssh_origin_vrf(self):
        from .machine import ssh_origin_vrf
        return self.machine.injector.get_instance(InjectionKey(ssh_origin_vrf, optional=True))


@inject(
    ainjector=AsyncInjector)
class SshKey(AsyncInjectable, SetupTaskMixin):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.config_layout = self.injector(ConfigLayout)

    @memoproperty
    def known_hosts(self):
        return self.config_layout.state_dir + "/ssh_known_hosts"

    async def async_ready(self):
        await self.run_setup_tasks()
        self.agent = await self.ainjector(ssh_agent, key=dependency_quote(self))
        return await super().async_ready()

    @setup_task('gen-key')
    async def generate_key(self):
        os.makedirs(self.config_layout.state_dir, exist_ok=True)
        await sh.ssh_keygen(f=self.key_path,
                            N='',
                            _in=None,
                            _bg=True,
                            _bg_exc=False)

    def add_to_agent(self, agent):
        try:
            sh.ssh_add(self.key_path, _env=agent.agent_environ)
        except sh.ErrorReturnCode:
            time.sleep(2)
            sh.ssh_add(self.key_path, _env=agent.agent_environ)

    @memoproperty
    def key_path(self):
        '''The path of the private key suitable for inclusion with ``ssh -i`` or *None* if no filesystem key exists.
        '''
        return self.config_layout.state_dir + '/ssh_key'

    @memoproperty
    def stamp_path(self):
        return self.config_layout.state_dir

    @memoproperty
    def ssh(self):
        return sh.ssh.bake(_env=self.agent.agent_environ)

    async def rsync(self, *args, ssh_origin=None):
        return await self.ainjector(
            rsync,
            *args,
            key=dependency_quote(self),
            ssh_origin=ssh_origin)

    @memoproperty
    def pubkey_contents(self):
        with open(self.key_path + ".pub", "rt") as f:
            return f.read()


@inject(
    key=InjectionKey(SshKey, _optional=True),
    config_layout=ConfigLayout,
    injector=Injector)
async def rsync(*args, config_layout,
                injector,
                ssh_origin=None, key=None):
    from .network import access_ssh_origin
    if key:
        ssh_agent = key.agent
    else:
        ssh_agent = injector.get_instance(SshAgent)
    ssh_options = []
    args = list(args)
    async with contextlib.AsyncExitStack() as stack:
        for i, a in enumerate(args):
            if isinstance(a, RsyncPath):
                if a.machine.rsync_uses_filesystem_access:
                    path = await stack.enter_async_context(a.machine.filesystem_access())
                    args[i] = Path(path).joinpath(a.relpath)
                    continue
                sso = a.ssh_origin
                if ssh_origin is None:
                    ssh_origin = sso
                    vrf = a.ssh_origin_vrf
                    ssh_options = list(a.machine.ssh_options) + a.machine.config_layout.global_ssh_options.split()
                elif ssh_origin is not sso:
                    raise RuntimeError(f"Two different ssh_origins: {sso} and {ssh_origin}")
                args[i] = str(a)
        if ssh_options:
            ssh_options = list(ssh_options)

        ssh_options.extend(['-oUserKnownHostsFile=' + config_layout.state_dir + "/ssh_known_hosts"])
        ssh_options = " ".join(ssh_options)
        rsync_opts = ('-e', 'ssh ' + ssh_options)

        if ssh_origin:
            return await access_ssh_origin(ssh_origin=ssh_origin,
                                           ssh_origin_vrf=vrf)(
                                               'rsync',
                                               *rsync_opts, *args,
                                               _bg=True, _bg_exc=False,
                                               _env=ssh_agent.agent_environ)
        else:
            return await sh.rsync(*rsync_opts, *args, _bg=True, _bg_exc=False,
                                  _env=ssh_agent.agent_environ)


@inject(
    ssh_key=SshKey,
    injector=Injector)
class AuthorizedKeysFile(Injectable):

    def __init__(self, ssh_key, injector):
        config_layout = injector(ConfigLayout)
        self.path = config_layout.state_dir + '/authorized_keys'
        authorized_keys = config_layout.authorized_keys
        if authorized_keys.startswith('|'):
            authorized_keys = authorized_keys[1:]
            keys_in = str(sh.sh("-c",
                                authorized_keys,
                                _encoding='utf-8'))
        else:
            if not authorized_keys:
                keys_in = ""
            else:
                keys_in = Path(authorized_keys).read_text()
        with open(self.path, "wt") as f:
            f.write(keys_in)
            if ssh_key:
                f.write(ssh_key.pubkey_contents)


@inject(
    injector=Injector,
    key=InjectionKey(SshKey, _optional=True, _ready=False))
class SshAgent(Injectable):

    def __init__(self, injector, key):
        config_layout = injector(ConfigLayout)
        run = config_layout.local_run_dir
        auth_sock = os.path.join(run, "ssh_agent")
        os.makedirs(run, exist_ok=True)
        try:
            os.unlink(auth_sock)
        except FileNotFoundError:
            pass
        if config_layout.production_ssh_agent and 'SSH_AUTH_SOCK' in os.environ:
            self.auth_sock = os.environ['SSH_AUTH_SOCK']
            self.process = None
        else:
            self.process = sh.ssh_agent('-a', auth_sock,
                                        '-D', _bg=True)
            self.auth_sock = auth_sock
        if key and is_obj_ready(key):
            self.handle_key(key)
        elif key:  # not ready
            future = asyncio.ensure_future(key.async_become_ready())
            future.add_done_callback(lambda f: self.handle_key(f.result()))

    def handle_key(self, key):
        key.add_to_agent(self)

    def close(self):
        if self.process is not None:
            try:
                self.process.terminate()
            except BaseException:
                pass
            self.process = None

    @memoproperty
    def agent_environ(self):
        env = os.environ.copy()
        env['SSH_AUTH_SOCK'] = self.auth_sock
        return env


ssh_agent = when_needed(SshAgent)

__all__ = ('SshKey', 'ssh_agent', 'SshAgent', 'RsyncPath', 'rsync')



# Copyright (C) 2019, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import os.path
from tempfile import TemporaryDirectory
from carthage import sh


class Machine:

    ssh_options = []

    def __init__(self, name):
        self.name = name
        self.dir = TemporaryDirectory()
        self.dir.__enter__()

    @property
    def ssh(self):
        environ = os.environ.copy()
        if 'PYTHONPATH' in environ:
            ppath_expanded = environ['PYTHONPATH'].split(':')
            ppath = ':'.join(map(lambda x: os.path.abspath(x), ppath_expanded))
            environ['PYTHONPATH'] = ppath

        class CommandFinder:

            def __call__(self, cmd, *args, **kwargs):
                cmd = getattr(self, cmd)
                return cmd(*args, **kwargs)

            def __getattr__(inner_self, c):
                c = getattr(sh, c)
                return c.bake(_cwd=self.dir.name,
                              _env=environ)
        return CommandFinder()

    def rsync_path(self, p):
        return os.path.join(self.dir.name,
                            os.path.relpath(p, start="/"))

    def close(self):
        self.dir.__exit__(None, None, None)

    def __del__(self):
        self.close()

    @property
    def path(self): return self.dir.name



# Copyright (C) 2019, 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage import *

from .utils import wait_for_task
from .folder import VmwareFolder, HostFolder
from .inventory import VmwareNamedObject
from .cluster import VmwareCluster
from pyVmomi import vim

__all__ = ['VmwareHost']


class VmwareHost(VmwareNamedObject, kind='host'):

    parent_type = (VmwareCluster, HostFolder)

    def __init__(self, name, **kwargs):
        if not name.startswith('/') and not 'parent' in kwargs:
            self.connection = kwargs['connection']  # for _find_by_name
            if not 'mob' in kwargs:
                kwargs['mob'] = self._find_by_name(name)
            kwargs['parent'] = self._parent_path_from_mob(kwargs['mob'].parent)
        kwargs['name'] = name
        self.spec = kwargs.pop('spec', None)
        super().__init__(**kwargs)

    def _find_parent(self):
        if (self.mob is not None) and isinstance(self.mob.parent, vim.ClusterComputeResource):
            self.parent_type = VmwareCluster
        else:
            self.parent_type = HostFolder
        return super()._find_parent()

    def _find_by_name(self, name):
        vm = self.connection.content.viewManager
        container = vm.CreateContainerView(self.connection.content.rootFolder, [vim.HostSystem], True)
        found = None
        for ref in container.view:
            if ref.name == name:
                found = ref
                break
        container.Destroy()
        return found

    async def do_create(self):
        task = self.parent.mob.AddHost_Task(spec=self.spec, asConnected=True)
        await wait_for_task(task)
        self.mob = task.info.result
        return self.mob



# Copyright (C) 2018, 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import contextlib
import pathlib
import os
import os.path
from .dependency_injection import *
from . import sh
from .config import ConfigLayout
from .utils import memoproperty
from . import machine
from .setup_tasks import *

@inject_autokwargs(config_layout=ConfigLayout)
class PkiManager(Injectable):

    def credentials(self, host):
        "Returns a key combined with certificate"
        self._certify(host)
        s = ""
        for ext in ('pem', 'key'):
            with open(os.path.join(self.pki_dir, "{}.{}".format(host, ext))) as f:
                s += f.read()
        return s

    def _certify(self, host):
        self.ca_cert
        sh.entanglement_pki(host, d=self.pki_dir, _bg=False)

    @memoproperty
    def pki_dir(self):
        ret = pathlib.Path(self.config_layout.state_dir)/"pki"
        ret.mkdir(exist_ok=True)
        return str(ret)

    @property
    def ca_cert(self):
        sh.entanglement_pki('-d', self.pki_dir,
                            '--ca-name', "Carthage Root CA", _bg=False)
        with open(self.pki_dir + '/ca.pem', 'rt') as f:
            return f.read()


__all__ = ['PkiManager']

def install_root_cert_customization(get_certificate_info):
    '''
    Return a customization that installs one or more root certificates.

    :param get_get_certificate_info: A callback function resolved in the context of the customization's :class:`AsyncInjector` that returns a sequence of [certificate_name, pem_certificate].  The *certificate_name* uniquely identifies the certificate and *pem_certificate* is the root certificate to add.

    '''
    class InstallRootCertCustomization(machine.FilesystemCustomization):

        @setup_task("Install  Root Certs")
        async def install_root_cert(self):
            certificates_to_install = await self.ainjector(get_certificate_info)
            carthage_cert_dir = os.path.join(
                    self.path,
                    "usr/share/ca-certificates/carthage")
            os.makedirs(carthage_cert_dir, exist_ok=True)
            for name, pem_cert in certificates_to_install:
                with open(os.path.join(
                        carthage_cert_dir, f"{name}.crt"),
                          "wt") as f:
                    f.write(pem_cert)
                with open(os.path.join(
                        self.path, "etc/ca-certificates.conf"),
                          "ta") as f:
                    f.write(f"carthage/{name}.crt\n")
            await self.run_command("/usr/sbin/update-ca-certificates")

    return InstallRootCertCustomization

__all__ += ['install_root_cert_customization']

@inject(
    pki=PkiManager)
def pki_manager_certificate_info(pki):
    return [('carthage_pki', pki.ca_cert)]

PkiCustomizations = install_root_cert_customization(pki_manager_certificate_info)

__all__ += ["PkiCustomizations"]



# Copyright (C) 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from __future__ import annotations
import argparse
from .console import CarthageRunnerCommand
from . import *
from . import kvstore
import asyncio

class MachineCommand(CarthageRunnerCommand):

    async def should_register(self):
        from .machine import Machine
        return any(self.ainjector.filter(Machine, ['host']))

    def setup_subparser(self, subparser):
        subparser.add_argument('machine')


class StartCommand(MachineCommand):

    name = 'start'

    async def run(self, args):
        machine = await self.ainjector.get_instance_async(InjectionKey(Machine, host=args.machine))
        if args.ready:
            if isinstance(machine.model, AsyncInjectable):
                await machine.model.async_become_ready()
            await machine.async_become_ready()
        await machine.start_machine()

    def setup_subparser(self, parser):
        parser.add_argument('--ready',
                            action=argparse.BooleanOptionalAction,
                            help='Bring machine to async_ready before starting')
        super().setup_subparser(parser)


class StopCommand(MachineCommand):

    name = 'stop'

    async def run(self, args):
        machine = await self.ainjector.get_instance_async(InjectionKey(Machine, host=args.machine))
        await machine.stop_machine()

class DeleteCommand(MachineCommand):

    name = 'delete'

    async def run(self, args):
        machine = await self.ainjector.get_instance_async(InjectionKey(Machine, host=args.machine))
        if hasattr(machine, 'delete'):
            await machine.delete()
        else:
            raise NotImplementedError(f'{machine} cannot be deleted')

class ListMachines(MachineCommand):

    name = 'list-machines'

    def setup_subparser(self, parser): pass

    async def run(self, args):
        machines = self.ainjector.filter(Machine, ['host'])
        for m in machines:
            print(m.host)

class SleepCommand(CarthageRunnerCommand):

    name = 'sleep'

    subparser_kwargs = dict(help='Disable the console command loop to prioritize PDB.')

    def setup_subparser(self, parser):
        pass

    async def run(self, args):
        # If no Pdb is active, this will return ~immediately.  If a
        # Pdb is active, this will yield the console until the Pdb is
        # continued or closed.
        await asyncio.sleep(0.01)

class DumpAssignmentsCommand(CarthageRunnerCommand):

    name = 'dump_assignments'

    subparser_kwargs = dict(
        help='Dump out assignments such as MAC addresses and IP addresses related to this layout.',
        )
    
    async def should_register(self):
        try:
            self.persistent_seed_path = await self.ainjector.get_instance_async(persistent_seed_path)
            return True
        except KeyError: return False

    def setup_subparser(self, parser):
        parser.add_argument('path',
                            default=self.persistent_seed_path,
                            nargs='?',
                            help=f'Path where assignments are dumped; defaults to {self.persistent_seed_path}')
        

    async def run(self, args):
        from carthage.modeling import CarthageLayout
        store = await self.ainjector.get_instance_async(kvstore.KvStore)
        layout = await self.ainjector.get_instance_async(CarthageLayout)
        models = await layout.all_models(ready=False)
        self.model_names = set((getattr(m, 'name',"") for m in models))
        store.dump(args.path, self.dump_filter)

    def dump_filter(self, domain, key, value):
        # Several domains have keys of the form model|interface
        if domain == 'mac' or domain.endswith('/hints') :
            name, sep, interface = key.partition('|')
            if sep:
                if name in self.model_names: return True
                return False
        # We do not recognize the structure so permit
        return True
    

def enable_runner_commands(ainjector):
    ainjector.add_provider(StartCommand)
    ainjector.add_provider(ListMachines)
    ainjector.add_provider(SleepCommand)
    ainjector.add_provider(StopCommand)
    ainjector.add_provider(DeleteCommand)
    ainjector.add_provider(DumpAssignmentsCommand)




# Copyright (C) 2019, 2020, 2021, 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import asyncio
import os
import pytest
import os.path
import sys
import shutil
from carthage.dependency_injection import *
from carthage.dependency_injection.introspection import *
from carthage_test_utils import Trigger
import carthage
import carthage.ansible
from carthage.pytest import *
from carthage.setup_tasks import *
from carthage.setup_tasks import SetupTaskContext, TaskWrapperBase
from pathlib import Path

state_dir = Path(__file__).parent.joinpath("test_state")


@pytest.fixture()
def ainjector(ainjector):
    ainjector = ainjector.claim("test_setup.py")
    config = ainjector.injector(carthage.ConfigLayout)
    config.state_dir = state_dir
    os.makedirs(state_dir, exist_ok=True)
    yield ainjector
    shutil.rmtree(state_dir, ignore_errors=True)


class Stampable(SetupTaskMixin, AsyncInjectable):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name = self.__class__.__qualname__

    def __repr__(self):
        return f"<setup_test object for {self.name} test>"

    def __init_subclass__(cls):
        super().__init_subclass__()
        cls.stamp_path = state_dir.joinpath(str(id(cls)))


@async_test
async def test_basic_setup(ainjector):
    called = 0

    class c(Stampable):
        @setup_task("test_stamp")
        def test_stamp_task(self):
            nonlocal called
            called += 1
    assert called == 0
    assert not c.check_stamp(c, "test_stamp_task")[0]
    assert len(list(c.inspect_class_setup_tasks())) == 1
    c_obj = await ainjector(c)
    assert called == 1
    assert c_obj.check_stamp("test_stamp_task")[0]
    inspectors = list(c_obj.inspect_setup_tasks())
    assert len(inspectors) == 1
    assert not await ainjector(inspectors[0].should_run)
    c_obj2 = await ainjector(c)
    assert c_obj is not c_obj2
    assert called == 1

@async_test
async def test_skip_task(ainjector):
    class c(Stampable):

        @setup_task("test_skip")
        def test_skip(self):
            raise SkipSetupTask

        @setup_task("Test async skip")
        async def test_async_skip(self):
            nonlocal async_called
            async_called = 1
            raise SkipSetupTask

    async_called = 0
    await ainjector(c)
    assert async_called == 1
        


@async_test
async def test_check_completed(ainjector):
    called = 0
    is_completed = True

    class c(Stampable):
        @setup_task("check_completed")
        def setup_check_completed(self):
            nonlocal called
            called += 1

        @setup_check_completed.check_completed()
        def setup_check_completed(self):
            return is_completed
    c_1 = await ainjector(c)
    assert called == 0
    is_completed = False
    c_2 = await ainjector(c)
    assert called == 1
    assert not c_1.check_stamp("check_completed")[0]


@async_test
async def test_invalidator(ainjector):
    called = 0

    class c(Stampable):
        @setup_task("test_invalidator")
        def setup_invalidator(self):
            nonlocal called
            called += 1

        @setup_invalidator.invalidator()
        def setup_invalidator(self, **kwargs):
            return False
    assert not c.check_stamp(c, "setup_invalidator")[0]
    await ainjector(c)
    assert called == 1
    assert c.check_stamp(c, "setup_invalidator")[0]
    await ainjector(c)
    assert called == 2


@async_test
async def test_failure_forces_rerun(ainjector):
    "If a task is explicitly run and fails, does the stamp get reset?"
    called = 0
    should_fail = False

    class c(Stampable):
        @setup_task("test_error_explicit")
        def setup_test_error_explicit(self):
            nonlocal called
            called += 1
            if should_fail:
                raise RuntimeError

    assert not c.check_stamp(c, "setup_test_error_explicit")[0]
    await ainjector(c)
    assert called == 1
    should_fail = True
    o = await ainjector(c)
    assert called == 1
    with pytest.raises(RuntimeError):
        o.setup_test_error_explicit()
    assert called == 2
    should_fail = False
    await ainjector(c)
    assert called == 3


@async_test
async def test_order_override(ainjector):
    two_called = False

    class c (Stampable):
        @setup_task("one")
        def one(self): pass

        @setup_task("foo", order=9200)
        def two(self):
            nonlocal two_called
            two_called = True

        @setup_task("three")
        def three(self):
            assert two_called is True

        @setup_task("before_two", before=two)
        def before_two(self):
            assert two_called is False

    assert c.one.order < 9200
    assert c.before_two.order < c.two.order
    assert c.three.order > c.two.order
    await ainjector(c)


@async_test
async def test_mako_task(ainjector):
    class bar(Stampable):

        templates = mako_task("test.mako")

        template_2 = mako_task("template-2.mako", name=InjectionKey('name'))

    ainjector.add_provider(InjectionKey("name"), "the name")
    res = await ainjector(bar)
    template_2 = res.stamp_path.joinpath("template-2").read_text()
    template_2_expected = state_dir.parent.joinpath("template-2.expected").read_text()
    assert template_2 == template_2_expected


@async_test
async def test_hash_func(ainjector):
    fake_hash = "30"
    test_hash_run = 0

    class c(Stampable):

        @setup_task("Test hash")
        def test_hash(self):
            nonlocal test_hash_run
            test_hash_run += 1

        @test_hash.hash()
        def test_hash(self):
            return fake_hash
    o = await ainjector(c)
    test_hash_run = 0
    o.test_hash()
    assert test_hash_run == 1
    await o.run_setup_tasks()
    assert test_hash_run == 1
    fake_hash = "45"
    await o.run_setup_tasks()
    assert test_hash_run == 2


@async_test
async def test_mako_hash(ainjector):
    class c(Stampable):

        mt = mako_task("test_hash.mako",
                       fake_hash=InjectionKey("fake_hash"),
                       real_value=InjectionKey("real_value"))

    def get_fake_hash(): return fake_hash
    def get_real_value(): return real_value
    ainjector.add_provider(InjectionKey("fake_hash"), get_fake_hash)
    ainjector.add_provider(InjectionKey("real_value"), get_real_value)
    real_value = "foo"
    fake_hash = "30"
    o = await ainjector(c)

    def output():
        return Path(o.stamp_path).joinpath("test_hash").read_text()
    assert output() == "foo"
    real_value = "bar"
    await o.run_setup_tasks()
    assert output() == "foo"
    fake_hash = "90"
    await o.run_setup_tasks()
    assert output() == "bar"


@async_test
async def test_setup_task_context(ainjector):
    class ContextTest(Stampable):

        @setup_task("Test context")
        async def test_context(self):
            nonlocal called
            ctx = current_instantiation()
            assert isinstance(ctx, SetupTaskContext)
            assert ctx.instance is self
            assert ctx.parent.key is InjectionKey(ContextTest)
            called = 1

    ainjector.add_provider(ContextTest)
    called = 0
    await ainjector.get_instance_async(ContextTest)
    assert called == 1


@async_test
async def test_add_setup_task(ainjector):
    def task_a(self):
        nonlocal task_a_called
        task_a_called = 1

    @setup_task("TaskWrapper test")
    def task_b(self):
        nonlocal task_b_called
        task_b_called = 1
    task_a_called = False
    task_b_called = False

    class C(Stampable):
        pass
    o = await ainjector(C)

    assert not task_a_called
    assert not task_b_called
    o.add_setup_task(task_a, description="call task_a")
    assert isinstance(task_b, TaskWrapperBase)
    o.add_setup_task(task_b)
    await o.run_setup_tasks()
    assert task_a_called
    assert task_b_called


@async_test
async def test_setup_task_introspection(ainjector):
    class Dependent(Stampable):

        @setup_task("Do some stuff in the dependent object")
        async def task_dependent(self):
            trigger1.trigger()
            await trigger2

    class Main(Stampable):

        @setup_task("Depends on setup task")
        @inject(d=Dependent)
        async def task_depending(self, d):
            await trigger3

    ainjector.add_provider(Dependent)
    ainjector.add_provider(Main)
    with Trigger() as trigger1, \
            Trigger() as trigger2, Trigger() as trigger3:
        main_future = asyncio.gather(ainjector.get_instance_async(Main))
        assert len(instantiation_roots) == 0
        await trigger1
        assert len(instantiation_roots) == 1
        icontext = next(iter(instantiation_roots))
        assert icontext.key == InjectionKey(Main)
        assert len(icontext.dependencies_waiting) == 1
        assert 'task_depending' in icontext.dependencies_waiting
        td_context = icontext.dependencies_waiting['task_depending']
        assert InjectionKey(Dependent) in td_context.dependencies_waiting
        trigger2.trigger()
        await ainjector.get_instance_async(Dependent)
        assert td_context._done is False
        trigger3.trigger()
        await main_future
        assert not td_context.dependencies_waiting
    assert not icontext.dependencies_waiting
    assert len(instantiation_roots) == 0
@async_test
async def test_setup_task_events(ainjector):
    class c(Stampable):

        @setup_task("Run some task")
        def task_1(self):
            pass

    def cb(event, target, **kwargs):
        events.add(event)
    events = set()
    c_not_ready = InjectionKey(c, _ready=False)
    ainjector.add_provider(c)
    c_obj = await ainjector.get_instance_async(c_not_ready)
    inspect = list(c_obj.inspect_setup_tasks())
    assert  await inspect[0].should_run(ainjector)
    with ainjector.event_listener_context(InjectionKey(c),
                                          ['task_run', 'task_should_run'], cb):
        await  c_obj.async_become_ready()
    assert 'task_should_run' in events
    assert 'task_run' in events
    assert 'task_already_run' not in events
    with ainjector.event_listener_context(
            InjectionKey(c), 'task_already_run', cb):
        c_2 = await ainjector(c)
    assert 'task_already_run' in events
    



# Copyright (C) 2022, 2023, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.
import collections.abc

import yaml
from pathlib import Path
from ipaddress import IPv4Address

import lmdb

from .dependency_injection import *
from .config import ConfigLayout
from .utils import memoproperty


__all__ = []

#: An injection key whose value is a path to be loaded into
#:class:`KvStore` at initialization time.  Typical usage is to
#periodically dump the dumpable domains (such as hints about IP
#address assignment) from the *KvStore* into a file checked into the
#source of a :class:`carthage.modeling.CarthageLayout`.  That layout
#then provides persistent_seed_path in the main injector of the
#layout.  At startup, the layout will load the hints file, so that
#multiple runs of the layout with different state directories will
#have consistent assignments.  Note that this must either be provided
#on the base injector or directly on a *CarthageLayout*.  There is
#special logic in *CarthageLayout* to perform the load when this key
#is set there.
persistent_seed_path = InjectionKey('carthage.kvstore/persistent_seed_path')

__all__ += ['persistent_seed_path']

@inject_autokwargs(config_layout=ConfigLayout,
                   persistent_seed_path=InjectionKey(persistent_seed_path, _optional=True),
                   )
class KvStore(Injectable):

    def __init__(
            self, store_dir="persistent_assignments", max_size=4*2**30,
            **kwargs):
        super().__init__(**kwargs)
        store_path = Path(store_dir)
        if not store_path.is_absolute():
            store_path = Path(self.config_layout.state_dir)/store_path
            Path(self.config_layout.state_dir).mkdir(parents=True, exist_ok=True)
        self.environment = lmdb.Environment(
            str(store_path), subdir=True,
            max_dbs=512,
            map_size=max_size,
            create=True,
            writemap=True)
        if self.persistent_seed_path:
            persistent_seed_path = Path(self.persistent_seed_path)
            if persistent_seed_path.exists():
                self.load(persistent_seed_path)
                

    def close(self):
        self.environment.close()

    def domain(self, d:str, include_in_dump):
        '''Return a :class:`KvDomain` for accessing a domain of keys in the Store.  Typical usage::

            kvstore = KvStore(path)
            domain = kvstore.domain('network/foo_net/addresses', True)
            domain.put('30', 'foo.com')   # foo.com is address 30 on this network

        :param include_in_dump: If True, then the contents of this domain should be included in the results of a call to :meth:`dump`
        '''
        if include_in_dump:
            with self.environment.begin(write=True) as txn:
                assert txn.put(b'dump:'+bytes(d, 'utf-8'), b'true', True)
        return KvDomain(self, d)

    def dump(self, file, filter):
        domains = set()
        file = Path(file)
        result = dict()
        with self.environment.begin() as txn, txn.cursor() as csr:
            csr.set_range(b'dump')
            for key, value in csr:
                if not key.startswith(b'dump:'): break
                domains.add(key[5:])
            for d in domains:
                domain_key = d+b':'
                csr.set_range(domain_key)
                domain_result = dict()
                for key, value in csr:
                    if not key.startswith(domain_key): continue
                    key = key[len(domain_key):]
                    key_str = str(key, 'utf-8')
                    value_str = str(value, 'utf-8')
                    if filter(str(d, 'utf-8'), key_str, value_str):
                        domain_result[key_str] = value_str
                result[str(d, 'utf-8')] = domain_result
        file.write_text(yaml.dump(result, default_flow_style=False))

    def load(self, file):
        '''Load a dump file produced by :meth:`dump`.  Values aready in the :class:`KvStore` override values in the dump.

        The intent is that a dump file can be checked into a layout repository as an initial set of assignments for things like IP addresses and MAC addresses.  An actual running state_dir for the layout may diverge from the initial hints contained in the dump.  Periodically the layout repository can be updated.
        '''
        file = Path(file)
        result = yaml.safe_load(file.read_text())
        with self.environment.begin(write=True) as txn:
            for domain, domain_dict in result.items():
                for k,v in domain_dict.items():
                    txn.put(kv_key(domain, k), bytes(v, 'utf-8'))
                    # Does not overwrite existing values
                    
    

__all__ += ['KvStore']

def kv_key(domain, key):
    domain = domain.replace(':', '::')
    return bytes(domain+':'+key, 'utf-8')


class KvDomain:

    def __init__(self, store, domain):
        self.domain = domain
        self.environment = store.environment

    def put(self, k, v, *,
            value=None,
            overwrite=False):
        '''Setself[*k* to *v*.  If *overwrite* is False, raise :class:`KvConsistency` if *k* in self and not self[*k*] == *value*.
        '''
        assert not ( value and overwrite), "Specifying overwrite and value nonsensical"
        key = kv_key(self.domain, k)
        v_bytes = bytes(v, 'utf-8')
        if value:
            value_bytes = bytes(value, 'utf-8')
        else: value_bytes = None
        with self.environment.begin(write=True) as txn:
            if value_bytes:
                actual_value = txn.get(key)
                if actual_value != value_bytes:
                    raise KvConsistency(f'Expecting {k} == {value} but actually {value_bytes}')
            if not txn.put(key, v_bytes, overwrite=(overwrite or value is not None)):
                raise KvConsistency(f'{k} present in {self.domain}')
            

    def get(self, k, default=None):
        '''Returns self[*k*] or if not present *default*'''
        key = kv_key(self.domain, k)
        with self.environment.begin() as txn:
            v = txn.get(key, NotPresent)
            if v == NotPresent: return default
            return str(v, 'utf-8')

    def delete(self, k, value=NotPresent):
        '''Removes *k* from self or raises :class:`KvConsistency`
        If *value* is given, then self[*k*] must equal *value* before the delete.
'''
        key = kv_key(self.domain, k)
        with self.environment.begin(write=True) as txn, \
             txn.cursor() as csr:
            csr.set_key(key)
            if csr.key() != key:
                raise KvConsistency(f'{k} not in {self.domain}')
            if value is not NotPresent:
                value_bytes = bytes(value, 'utf-8')
                if csr.value() != value_bytes:
                    raise KvConsistency(f'{k} in {self.domain} had unexpected value')
            csr.delete()


    def __getitem__(self, k):
        v = self.get(k, NotPresent)
        if v is NotPresent:
            raise KeyError(k)
        return v

    def __delitem__(self, k):
        try: self.delete(k)
        except KvConsistency:
            raise KeyError(k) from None
        

class KvConsistency(RuntimeError):
#    def __init__(self, *args):
#        breakpoint()

    pass


__all__ += ['KvConsistency']


class AssignmentsExhausted(RuntimeError): pass

__all__ += ['AssignmentsExhausted']


@inject_autokwargs(
    store=KvStore,
    )
class HintedAssignments(Injectable):

    #: How many times to retry an assignment when we lose a race against another process.
    consistency_retries = 5

    def __init__(self, domain, **kwargs):
        super().__init__(**kwargs)
        self._assignments = self.store.domain(domain+'/assignments', False)
        self._hints = self.store.domain(domain+'/hints', True)
        self._can_validate_assignments = False
        self.prefer_reallocate = False #: move things around when the preferred assignment changes
        self.new_assignments()


    def new_assignments(self):
        "Indicate a new round of assignments is beginning.  The same assignment will never be used more than once in a single round of assignments, but may be reused across rounds."
        self._assignments_made = dict()

    def _assign(self,  key, obj):
        '''
Called in subclasses to indicate that an assignment should be made.  :meth:`record_assignment` is called to let the object know what its assignment is.

        :param key: The key under which an assignment or hint is registered.  Should be unique across runs for the same object.

        :param obj: The object corresponding to *key*.  Not used by :class:`HintedAssignments` except as an input to the subclass's :meth:`record_assignment`
        '''
        for i in range(self.consistency_retries):
            try:
                hint = self._hints.get(key)
                if hint and self.valid_assignment(hint, obj):
                    # We always try to reuse a hint
                    if self._try_assignment(key, obj, hint, True):
                        return
                    else:
                        try: self._hints.delete(key, value=hint)
                        except KvConsistency:
                            logger.debug(f'Tried deleting hint for {key} but it was not {hint}')
                # No hint
                reusable_assignment = None
                for assignment in self.possible_assignments(key, obj):
                    result = self._try_assignment(key, obj, assignment, self.prefer_reallocate)
                    if result is True: return
                    if result == "reusable":
                        # If we preferred reusing an assignment, then we already would have done so.
                        # Remember the first reusable assignment and use if all assignments are exhausted.
                        if reusable_assignment is None: reusable_assignment = assignment
                if reusable_assignment:
                    if self._try_assignment(key, obj, reusable_assignment, True):
                        return
                raise AssignmentsExhausted(f'Assignments for {self} exhausted')
                
            except KvConsistency:
                continue
        raise KvConsistency(f'Exceeded maximum retries')


    def _try_assignment(self, key, obj, assignment, reallocate_assigned):
        '''
        Called to  try and assign *assignment* as the assignment for *key*.  This can work either if:

        * *assignment* is unassigned in the store of assignments

        *Or *assignment* is assigned to a key for which :meth:`valid_key` returns False.  In this case the assignment is reusable.

          If an assignment is reusable and *reallocate_assign* is True, then the assignment is reallocated.  If *reallocate_assign* is False, then ``'reuseable'`` is returned.

        :returns: True if the assignment is made; False if the requested assignment is allocated to a valid key; or ``'reusable'`` if the assignment could be reallocated but *reallocate_assignment* is False.

        '''
        try: del self._assignments_made[key]
        except KeyError: pass
        current_key = self._assignments.get(str(assignment))
        if current_key in self._assignments_made and self._assignments_made[current_key] == str(assignment):
            return False   # Has been allocated in this round to someone else
        elif current_key in self._assignments_made:
            # Has been assigned a different assignment this round
            try: self._assignments.delete(key, value=current_key)
            except KvConsistency: return False
            # If we did delete the assignment, then we can use the
            # key.  Don't count as a reallocate, because the previous
            # allocated key has already moved.
            current_key = None
        if current_key and (current_key != key):   # Potentially allocated see if still valid
            if not self._can_validate_assignments: return False   # Models not loaded enough to know what all the valid keys are
            if self.valid_key(current_key):
                return False   # Allocated to valid key
            # It's reusable
            if not reallocate_assigned: return 'reusable'
        # Is available for us to assign to key
        self._assignments.put(str(assignment), key, value=current_key)
        # Past this point we should not get KvConsistency errors.
        self._hints.put(key, str(assignment), overwrite=True)
        # There is a race.  By recording the assignment before the hint, if
        # someone else is also assigning the same object at the same time,
        # they may not get the hint and may try enumerating possible
        # assignments.  Assuming possible_assignments is stable, they will end
        # up with the same result.  If not, it is possible there could be
        # churn.  By recording the hint as soon as we can, we minimize this.
        # We could avoid this with some sort of multiput if it allowed us to
        # require a value for one key but not the other.  Redis doesn't appear
        # to have that.  Based on expected usage, we accept the race.
        self.record_assignment(key, obj, assignment)
        self._assignments_made[key] = str(assignment)
        return True

    def force_assignment(self, key, obj, assignment):
        '''Force recording within the data store that *obj* identified by *key* has *assignment* as its assignment.  This is intended for dealing with statically assigned assignments that fall into the range that is automatically managed.  Does not call :meth:`record_assignment`
        '''
        self._assignments.put(key, str(assignment), overwrite=True)
        self._hints.put(key, str(assignment), overwrite=True)
        self._assignments_made[key] = str(assignment)
        
    def record_assignment(self, key, obj, assignment):
        '''
        Implemented by subclasses to inform an object of its assignment.
        As an example if being used for IP address asignment, the subclass version of this method would be responsible for setting the address on the :class:`carthage.network.NetworkLink`.
        Note that *assignment* may be a string, even if :meth:`possible_assignments` yields something like integers.  This will happen for example when a potential assignment is retrieved from  a hint.
        '''
        raise NotImplementedError


    def possible_assignments(self, key, obj):
        '''Implemented by subclasses.  This generator yields all the possible assignments for a given *obj* in preference order.
It is best if this generator yields strings; ass discussed in the documentation for :meth:`record_assignment` at least for hints the assignment code will coerce assignments to strings.
'''
        raise NotImplementedError

    
    def enable_key_validation(self):
        '''Called by the subclass when valid_key can be called.  Prior to this method being called, valid_key will never be called under the assumption that not all valid keys are known.  So prior to this method being called, no assignments can be reallocated.
        '''
        self._can_validate_assignments = True

    def valid_assignment(self, assignment, obj):
        '''
        Implemented by subclass. Return True if *assignment* is still a valid assignment for this collection.  Used to confirm a hint is reasonable and to cleanup assignments that are no longer possible.
        If the possibility of an illegal assignment still being hinted is not an issue even when the set returned by possible_assignments changes, this method can always return True.
        '''
        return True

    def valid_key(self, key):
        '''
        Implemented by subclass. Returns True if *key* represent an object that we still want to track assignments for. If this returns False then resources assigned to *key* may be reallocated for other keys.
        '''
        raise NotImplementedError


__all__ += ['HintedAssignments']


class HashedRangeAssignments(HintedAssignments):

    def __init__(self, domain,  **kwargs):
        super().__init__(domain, **kwargs)

    

    def hash_key(self, key, obj):
        '''Key hashed, bounded to low <= key <= high
:returns: low, hash, high
'''
        assert isinstance(key, str)
        low, high = self.find_bounds(obj)
        result = 0
        for c in key: result += ord(c)
        try: size = high-low +1
        except TypeError:
            size = int(high)-int(low)+1
        return low, low + (result % size), high

    

    def possible_assignments(self, key, obj):
        '''Finds the bounds for *obj* using :meth:`find_bounds` then returns all assignments within the bounds:

        * Initially ``self.hash_key())

        * If that doesn't work, try at a distance of 1, rejecting either ``hash+1`` or ``hash-1`` if it falls outside the bounds
       
        * Increase distance, stopping iteration if both ``hash+distance`` and ``hash-distance`` are out of bounds.

        '''
        low, high = self.find_bounds(obj)
        low, hash, high = self.hash_key(key, obj)
        result_yielded = True
        distance = 0
        while result_yielded:
            result_yielded = False
            if low <= (hash+distance) <= high:
                yield str(hash+distance)
                result_yielded = True
            if (distance > 0) and (low <= hash-distance <= high):
                yield str(hash-distance)
                result_yielded = True
            distance += 1

    def find_bounds(self, obj):
        '''Implemented by subclass; returns tuple of low, high, for the range of assignments for this object.
        '''
        raise NotImplementedError

    def valid_assignment(self, assignment, obj):
        assignment = self.str_to_assignment(assignment)
        low, high = self.find_bounds(obj)
        if low <= assignment <= high: return True
        return False

    def str_to_assignment(self, s):
        '''Override if needed to convert a string into an assignment useful for valid_assignment
'''
        return int(s)

__all__ += ['HashedRangeAssignments']




# Copyright (C) 2019, 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

import os.path
from ..dependency_injection import inject, InjectionKey, Injectable, Injector
from .layout import ConfigLayout


def getattr_path(o, attrs):
    attrs_iter = attrs
    try:
        while attrs_iter:
            left, sep, attrs_iter = attrs_iter.partition('.')
            o = getattr(o, left)
            if not attrs_iter:
                return o
    except AttributeError:
        raise AttributeError(f'Unable to find {attrs}') from None


@inject(
    injector=Injector)
class ConfigString(str):

    '''A string that substitutes

    * ``{key}`` with the result of that config key

    * ``{plugin:selector}`` with the lookup of *selector* using the *plugin* :ref:`.ConfigLookupPlugin`

    Backslash scapes the next character always; ``${var}`` is reserved for :ref:`ConfigPath` use.

    '''

    @classmethod
    def parse(cls, s, config, injector):
        def tok(i, awaiting_brace):
            # takes an iterator to consume input characters
            lastch = None
            for c in i:
                if lastch == '\\':
                    yield c
                elif c == '\\':
                    # one backslash always eats itself.
                    pass
                elif c == '{':
                    if lastch == "$":
                        yield '{'
                        yield from tok(i, True)
                        yield "}"
                    else:
                        yield from iter(cls.subst(
                            "".join(tok(i, True)),
                            config=config,
                            injector=injector
                        ))
                elif c == '}':
                    if awaiting_brace:
                        return  # end of inner token
                    else:
                        raise ValueError(f"Unbalanced closing brace in `{s}'")
                else:
                    yield c
                lastch = c

            if awaiting_brace:
                raise ValueError(f"Missing right brace in `{s}'")

        return "".join(tok(iter(s), False))

    @staticmethod
    def subst(s, *, config, injector):
        plugin, sep, selector = s.partition(':')
        if sep == '':
            return getattr_path(config, s)
        else:
            try:
                plugin = injector.get_instance(InjectionKey(ConfigLookupPlugin, name=plugin))
            except KeyError:
                raise KeyError(f'Config lookup plugin {plugin} not found') from None
            return plugin(selector)

    def __new__(cls, s, *, injector):
        config = injector(ConfigLayout)
        return str.__new__(str, cls.parse(os.path.expandvars(s), config, injector))


@inject(
    injector=Injector)
class ConfigPath(ConfigString):

    def __new__(cls, s, *, injector):
        return super().__new__(ConfigString, os.path.expanduser(s),
                               injector=injector)


class ConfigBool:

    "A type that can be subtyped to be injectable used instead of bool"

    def __new__(cls, val):
        return bool(val)


class ConfigLookupPlugin (Injectable):

    '''
    An abstract class representing an interface that can be used to look up information in an external store to fill in config values.  Used in config schemas like::

        password: {plugin:where_to_find_password}

    Registered like::

        ConfigPasswordPlugin.register(base_injector, "password")

'''

    @classmethod
    def register(cls, injector, name):
        injector.add_provider(InjectionKey(ConfigLookupPlugin, name=name), cls)

    def __call__(self, selector):
        '''
        return the value looked up using the given selector.
Must be overridden; this is abstract.
'''



# Copyright (C) 2019, 2020, Hadron Industries, Inc.
# Carthage is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License version 3
# as published by the Free Software Foundation. It is distributed
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the file
# LICENSE for details.

from carthage import *

import datetime
import time
import os
import types

from pyVmomi import vim, vmodl

from .connection import VmwareConnection
vmware_config = config_key('vmware')
__all__ = "VmwareStampable VmwareManagedObject VmwareNamedObject VmwareSpecifiedObject custom_fields_key default_custom_fields all_objs".split()

#: The provider for this key should be a dictionary mapping field names to functions.  The function takes a VmwareStampable and returns the field value.
custom_fields_key = InjectionKey('vmware.custom_fields_key')


@inject(config=ConfigLayout)
def vmware_dict(config, **kws):
    '''
:returns: A dictionary containing vmware common parameters to pass into Ansible
'''
    vconfig = config.vmware
    d = dict(
        datacenter=vconfig.datacenter,
        username=vconfig.username,
        hostname=vconfig.hostname,
        validate_certs=vconfig.validate_certs,
        password=vconfig.password)
    d.update(kws)
    return d


class NotFound(LookupError):
    pass


@inject_autokwargs(config_layout=ConfigLayout,
                   injector=Injector,
                   connection=VmwareConnection)
class VmwareStampable(SetupTaskMixin, AsyncInjectable):

    def __init_subclass__(cls, *, kind=NotImplemented):
        if kind is not NotImplemented:
            cls.stamp_type = kind

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @memoproperty
    def stamp_descriptor(self):
        raise NotImplementedError(type(self))

    @memoproperty
    def stamp_path(self):
        p = self.config_layout.state_dir
        p = os.path.join(p, "vmware_stamps", self.stamp_type)
        p = os.path.join(p, *self.stamp_descriptor.lstrip('/').split('/'))
        p += ".stamps"
        os.makedirs(p, exist_ok=True)
        return p


class VmwareManagedObject(VmwareStampable):

    '''Contains a reference to a VMware managed object.'''

    # custom fields
    created = 'com.hadronindustries.carthage.created'

    def __init__(self, *args, parent=None, mob=None, readonly=None, **kwargs):
        '''

        :param parent: the parent Carthage object or alternatively an inventory path of the parent.
        :param mob: manage an existing VMware managed object
        :param readonly: set if it should be readonly; should exist already; should not be removed or changed

        '''

        if isinstance(parent, str):
            self.parent = None
            self.parent_path = self.canonicalize_path(parent)
        else:
            if not isinstance(parent, self.parent_type):
                raise TypeError(f'{type(parent)} is not a valid type to be a parent of {type(self)}')
            self.parent = parent
            if parent is not None:
                self.parent_path = self.parent.vmware_path
        if self.is_root:
            self.parent_path = ""
        if not hasattr(self, 'parent_path'):
            raise TypeError(f"Parent must be set for {type(self)}")
        # If you need to be able to pass in mob without parent or parent path, I
        # know how to do that but it's not yet implemented; ask
        self.mob = mob
        if self.parent and self.mob:
            assert self.mob.parent == self.parent.mob

        self.writable = None if readonly is None else not readonly
        super().__init__(*args, **kwargs)

    @memoproperty
    def stamp_descriptor(self):
        return self.vmware_path

    #: False if this object should have a parent
    is_root = False

    @setup_task("Construct object")
    async def construct(self):
        if not self.mob:
            if not self.writable:
                raise NotFound(f'{type(self)} with path {self.vmware_path} does not exist')
        await self.do_create()
        self.mob = self._find_from_path()
        if self.mob is None:
            raise RuntimeError(f'constructed object, but could not find it at {self.vmware_path}')
        self.set_custom_fields()

    @construct.check_completed()
    async def construct(self):
        if not self.mob:
            self.mob = self._find_from_path()
        await self._find_parent()
        if not self.mob:
            return False
        v = self.get_field_value(self.created)
        if v is None:
            return True  # We don't know the dependency
        return datetime.datetime.fromisoformat(v).timestamp()

    async def _find_parent(self):
        if self.parent or self.is_root:
            return
        parent_key = InjectionKey(self.parent_type, path=self.parent_path)
        try:
            self.parent = await self.ainjector.get_instance_async(parent_key)
            return
        except KeyError:
            connection_injector = self.ainjector.injector_containing(VmwareConnection)
            connection_injector.add_provider(parent_key, when_needed(self.parent_type, name=self.parent_path))
            self.parent = await self.ainjector.get_instance_async(parent_key)

    async def do_create(self):
        raise NotImplementedError

    def children(self, objtypes, recursive=True):
        if self.mob is None:
            return
        vm = self.connection.content.viewManager
        container = None
        try:
            container = vm.CreateContainerView(self.mob, objtypes, recursive)
            for ref in container.view:
                yield ref
        finally:
            if container is not None:
                container.Destroy()

    @staticmethod
    def canonicalize_path(path):
        parts = [x for x in path.split('/') if x != '']
        return '/' + '/'.join(parts)

    @memoproperty
    def vmware_path(self):
        raise NotImplementedError

    async def get_permissions(self):
        am = self.connection.content.authorizationManager
        return am.RetrieveEntityPermissions(self.mob, inherited=True)

    async def add_permissions(self, permissions):
        raise NotImplementedError

    async def set_permissions(self, permissions):
        am = self.connection.content.authorizationManager
        vmware = [vim.AuthorizationManager.Permission(
            entity=self.mob,
            principal=permission.principal.principal,
            group=permission.principal.group,
            roleId=permission.role.mob.roleId,
            propagate=permission.propagate
        )
            for permission in permissions]
        am.SetEntityPermissions(self.mob, vmware)

    async def _find_by_name(self, name, vimtype):

        vm = self.connection.content.viewManager
        container = vm.CreateContainerView(self.connection.content.rootFolder, [vimtype], True)
        found = None
        for ref in container.view:
            if ref.name == name:
                found = ref
                break
        container.Destroy()
        return found

    def _find_from_path(self):
        find = self.connection.content.searchIndex.FindByInventoryPath
        ret = find(self.vmware_path)
        return ret

    def set_custom_fields(self):
        entity = self.mob
        if not self.writable:
            return
        fields = self.injector.get_instance(custom_fields_key)
        for name, val_func in fields.items():
            field = self._ensure_custom_field(name, vim.ManagedEntity)
            self.set_custom_field(field, val_func(self))

    def _fetch_custom_field(self, fname):
        content = self.connection.content
        cfm = content.customFieldsManager
        for f in cfm.field:
            if f.name == fname:
                return f
        raise KeyError(fname)

    def _ensure_custom_field(self, fname, ftype):
        try:
            return self._fetch_custom_field(fname)
        except KeyError:
            content = self.connection.content
            cfm = content.customFieldsManager
            return cfm.AddFieldDefinition(name=fname, moType=ftype)

    def set_custom_field(self, field, value):
        if self.mob is None:
            raise RuntimeError('unable to set fields on null object')
        if isinstance(field, str):
            field = self._fetch_custom_field(field)
        content = self.connection.content
        cfm = content.customFieldsManager
        cfm.SetField(entity=self.mob, key=field.key, value=value)

    def get_field_value(self, field):
        '''Return the vmware custom field value or None if not set
        '''
        if isinstance(field, str):
            field = self._ensure_custom_field(field, vim.ManagedEntity)

        for val in self.mob.customValue:
            if (val.key == field.key) and (val.value != ''):
                return val.value
        return None

    @staticmethod
    def _parent_path_from_mob(mob):
        parts = []
        while mob:
            parts.append(mob.name)
            if isinstance(mob, vim.Datacenter):
                mob = None
            else:
                mob = mob.parent
        return "/" + "/".join(reversed(parts))


class VmwareNamedObject(VmwareManagedObject):

    def __init__(self, name=None, *args, **kwargs):
        parent = kwargs.get('parent', None)
        if name is None:
            raise ValueError(f'must specify name')
        if parent and name.startswith('/'):
            raise TypeError("Cannot specify both a parent and a name containing a full path")
        elif parent and '/' in name:
            parent_add, sep, name = name.rpartition('/')
            if not isinstance(parent, str):
                parent = parent.vmware_path
            kwargs['parent'] = parent + '/' + parent_add
        elif parent is None and '/' in name:
            kwargs['parent'], sep, name = name.rpartition('/')
        if 'mob' in kwargs and name is None:
            name = kwargs['mob'].name
        #: The name of the object
        self.name = name
        super().__init__(*args, **kwargs)

    @memoproperty
    def vmware_path(self):
        return f'{self.parent_path}/{self.full_name}'

    @property
    def full_name(self):
        '''Returns the full name of the object.  The typical difference between this and name is that full_name includes the container_prefix for objects where that is necessary
'''
        return self.name

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.vmware_path}>"


class VmwareSpecifiedObject(VmwareNamedObject):

    #: Class variable containing set of `ConfigSpecStages` for this type of object.  Copied on any subclass that has *ConfigSpecStages*
    config_stages = []

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    async def build_config(self, mode, oconfig=None, stagefilter=None):
        if stagefilter is None:
            def stagefilter(cs): return mode in cs.mode
        ainjector = self.ainjector
        bag = types.SimpleNamespace(mode=mode)
        stages = []
        for cs in self.__class__.config_stages:
            if not stagefilter(cs):
                continue
            stages.append(cs(obj=self, bag=bag))
        config = self.config_spec_class()
        if self.mob:
            config.version = self.mob.config.version
        for s in stages:
            if oconfig:
                s.oconfig = oconfig
            await ainjector(s.apply_config, config)
        return config


def all_objs(content, root, objtype):
    vm = content.viewManager
    container = vm.CreateContainerView(root, objtype, True)
    for ref in container.view:
        yield ref
    container.Destroy()


def created_current_time(_):
    return datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()


default_custom_fields = {
    VmwareManagedObject.created: created_current_time}
